﻿#define DEBUG

using System;
using System.Windows.Forms;
using Core.Controllers.Gamepad;
using Core.Controllers.Gamepad.Enums;
using Core.Controllers.Helpers;
using Core.Debug;
using Core.Model;
using Core.Model.Chunks;
using Core.Model.Types;

namespace Core.Controllers
{
    public class Camera
    {
        private const float MOVEMENTSPEED = 0.5f;
        private const float ROTATIONSPEED = 0.1f;

        public MatrixF View;
        public MatrixF Projection;
        public MatrixF World = MatrixF.Identity;

        public static readonly int MaxYPosition = Chunk.Size.Y + 5;

        private float _viewAngle = MathHelper.PiOver2;
        private float _nearPlane = 0.01f;
        private float _farPlane = 10000f;

        private MouseState _mouseMoveState;

        private float _leftRightRotation = 0f;
        private float _upDownRotation = 0f;

        private Vector3F _cameraFinalTarget;
        private Vector3F _lookVector;

        public Vector3F Position = new Vector3F(10000, 128, 10000);
        public Vector3I PositionIndex;

        public int ViewPortWidth;
        public int ViewPortHeight;

        private Keyboard Keyboard;

        private Mouse Mouse;

        private WorldDebugger Debugger;

        public World CameraWorld;

        #region ctor

        public Camera()
        {
            OnPositionChanged();
        }

        #endregion

        #region Public methods

        public void InitializeCamera(World world)
        {
            CameraWorld = world;

            Debugger = world.Debugger;

            Keyboard = new Keyboard();

            Mouse = new Mouse(world.Render.WindowHandle);

            ViewPortWidth = world.Render.WindowWidth;
            ViewPortHeight = world.Render.WindowHeight;

            CalculateProjection((float)ViewPortWidth / (float)ViewPortHeight);

            CalculateView();
        }

        public void Update()
        {
            CheckKeyboard(Keyboard);

            CheckMouse(Mouse);
        }

        #endregion

        #region Protected methods

        protected void CalculateProjection(float aspectRatio)
        {
            Projection = MatrixF.CreatePerspectiveFieldOfView(_viewAngle, aspectRatio, _nearPlane, _farPlane);
        }

        protected void CalculateView()
        {
            MatrixF rotationMatrix = MatrixF.CreateRotationX(_upDownRotation) *
                                    MatrixF.CreateRotationY(_leftRightRotation);

            _lookVector = Vector3F.Transform(Vector3F.Forward, rotationMatrix);

            _cameraFinalTarget = Position + _lookVector;

            Vector3F cameraRotatedUpVector = Vector3F.Transform(Vector3F.Up, rotationMatrix);
            View = MatrixF.CreateLookAt(Position, _cameraFinalTarget, cameraRotatedUpVector);
        }

        #endregion

        #region Private methods

        private void OnPositionChanged()
        {
            PositionIndex = new Vector3I((uint)(Position.X / Chunk.Size.X), 0, (uint)(Position.Z / Chunk.Size.Z));
        }

        private void CheckKeyboard(Keyboard keyboard)
        {
            Vector3F moveVector = Vector3F.Zero;

            if (keyboard.IsKeyPressed(Keys.Escape))
            {
                Application.Exit();
            }

            if (keyboard.IsKeyPressed(Keys.W))
            {
                moveVector += Vector3F.Forward;
            }

            if (keyboard.IsKeyPressed(Keys.S))
            {
                moveVector += Vector3F.Backward;
            }

            if (keyboard.IsKeyPressed(Keys.A))
            {
                moveVector += Vector3F.Left;
            }

            if (keyboard.IsKeyPressed(Keys.D))
            {
                moveVector += Vector3F.Right;
            }

            MoveTo(moveVector);

            #if DEBUG

            if (Debugger != null)
            {
                Debugger.OnCheckKeyboard(keyboard);
            }

            #endif
        }

        private void CheckMouse(Mouse mouse)
        {
            MouseState currentMouseState = mouse.GetState();

            RotateCamera(currentMouseState.X - _mouseMoveState.X, currentMouseState.Y - _mouseMoveState.Y);

            _mouseMoveState = new MouseState(ViewPortWidth / 2, ViewPortHeight / 2,
                                             0, MouseButtonState.Released,
                                             MouseButtonState.Released, MouseButtonState.Released,
                                             MouseButtonState.Released, MouseButtonState.Released);

            mouse.SetPosition(_mouseMoveState.X, _mouseMoveState.Y);
            _mouseMoveState = mouse.GetState();
        }

        private void RotateCamera(float mouseDX, float mouseDY)
        {
            if (mouseDX != 0)
            {
                _leftRightRotation -= ROTATIONSPEED*(mouseDX/50);
            }
            else if (mouseDY == 0)
            {
                return;
            }

            if (mouseDY != 0)
            {
                _upDownRotation -= ROTATIONSPEED*(mouseDY/50);

                // Locking camera rotation vertically between +/- 180 degrees
                float newPosition = _upDownRotation - ROTATIONSPEED*(mouseDY/50);

                if (newPosition < -1.55f)
                {
                    newPosition = -1.55f;
                }
                else if (newPosition > 1.55f)
                {
                    newPosition = 1.55f;
                }

                _upDownRotation = newPosition;
            }
            else if (mouseDX == 0)
            {
                return;
            }

            CalculateView();
        }

        private void MoveTo(Vector3F moveVector)
        {
            if (moveVector != Vector3F.Zero)
            {
                MatrixF rotationMatrix = MatrixF.CreateRotationX(_upDownRotation)*
                                        MatrixF.CreateRotationY(_leftRightRotation);

                Vector3F rotatedVector = Vector3F.Transform(moveVector, rotationMatrix);

                Position += rotatedVector*MOVEMENTSPEED;

                if (Position.Y > MaxYPosition)
                {
                    Position = new Vector3F(Position.X, MaxYPosition, Position.Z);
                }

                OnPositionChanged();

                CalculateView();
            }
        }

        #endregion
    }
}
